@@ -6,8 +6,9 @@
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
-   License as published by the Free Software Foundation; either
-   version 2 of the License, or (at your option) any later version.
+   License as published by the Free Software Foundation, with
+   linking exception; either version 2.1 of the License, or (at
+   your option) any later version.
 
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -16,15 +17,12 @@
 
    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *)
 
-let rec first f l =
-  match l with
-    []     -> None
-  | x :: r -> match f x with
-                None          -> first f r
-              | Some _ as res -> res
+module Cset = Re_cset
+module Automata = Re_automata
+module MarkSet = Automata.PmarkSet
 
 let rec iter n f v = if n = 0 then v else iter (n - 1) f (f v)
 
@@ -33,10 +31,19 @@ let rec iter n f v = if n = 0 then v else iter (n - 1) f (f v)
 let unknown = -2
 let break = -3
 
-type 'a match_info =
-  [ `Match of 'a
-  | `Failed
-  | `Running ]
+(* Result of a successful match. *)
+type groups = {
+  s : string ;
+  marks : Automata.mark_infos ;
+  pmarks : MarkSet.t ;
+  gpos : int array ;
+  gcount : int
+}
+
+type match_info =
+  | Match of groups
+  | Failed
+  | Running
 
 type state =
   { idx : int;
@@ -51,7 +58,7 @@ type state =
         (* Transition table, indexed by color *)
     mutable final :
       (Automata.category *
-       (Automata.idx * Automata.mark_infos match_info)) list;
+       (Automata.idx * Automata.status)) list;
         (* Mapping from the category of the next character to
            - the index where the next position should be saved
            - possibly, the list of marks (and the corresponding indices)
@@ -65,9 +72,9 @@ type re =
         (* The whole regular expression *)
     mutable initial_states : (int * state) list;
         (* Initial states, indexed by initial category *)
-    cols : string;
+    cols : Bytes.t;
         (* Color table *)
-    col_repr : string;
+    col_repr : Bytes.t;
         (* Table from colors to one character of this color *)
     ncol : int;
         (* Number of colors *)
@@ -81,13 +88,15 @@ type re =
     group_count : int
         (* Number of groups in the regular expression *) }
 
-let print_re ch re = Automata.print_expr ch re.initial
+let pp_re ch re = Automata.pp ch re.initial
+
+let print_re = pp_re
 
 (* Information used during matching *)
 type info =
   { re : re;
         (* The automata *)
-    i_cols : string;
+    i_cols : Bytes.t;
         (* Color table ([x.i_cols = x.re.cols])
            Sortcut used for performance reasons *)
     mutable positions : int array;
@@ -98,6 +107,7 @@ type info =
     mutable last : int
         (* Position where the match should stop *) }
 
+
 (****)
 
 let cat_inexistant = 1
@@ -111,8 +121,10 @@ let category re c =
   if c = -1 then cat_inexistant else
   (* Special category for the last newline *)
   if c = re.lnl then cat_lastnewline lor cat_newline lor cat_not_letter else
-  match re.col_repr.[c] with
-    'a'..'z' | 'A'..'Z' ->
+  match Bytes.get re.col_repr c with
+    (* Should match [cword] definition *)
+    'a'..'z' | 'A'..'Z' | '0'..'9' | '_' | '\170' | '\181' | '\186'
+  | '\192'..'\214' | '\216'..'\246' | '\248'..'\255' ->
       cat_letter
   | '\n' ->
       cat_not_letter lor cat_newline
@@ -128,12 +140,12 @@ let unknown_state =
     next = dummy_next; final = [];
     desc = Automata.dummy_state }
 
-let count = ref 0
 let mk_state ncol ((idx, _, _, _, _) as desc) =
   let break_state =
     match Automata.status desc with
-      `Running -> false
-    | _        -> true
+    | Automata.Running -> false
+    | Automata.Failed
+    | Automata.Match _ -> true
   in
   { idx = if break_state then break else idx;
     real_idx = idx;
@@ -161,8 +173,8 @@ let delta info cat c st =
   end;
   desc
 
-let validate info s pos st =
-  let c = Char.code info.i_cols.[Char.code s.[pos]] in
+let validate info (s:string) pos st =
+  let c = Char.code (Bytes.get info.i_cols (Char.code s.[pos])) in
   let cat = category info.re c in
   let desc' = delta info cat c st in
   let st' = find_state info.re desc' in
@@ -187,9 +199,9 @@ let rec loop info s pos st =
     st
 *)
 
-let rec loop info s pos st =
+let rec loop info (s:string) pos st =
   if pos < info.last then
-    let st' = st.next.(Char.code info.i_cols.[Char.code s.[pos]]) in
+    let st' = st.next.(Char.code (Bytes.get info.i_cols (Char.code s.[pos]))) in
     loop2 info s pos st st'
   else
     st
@@ -202,7 +214,7 @@ and loop2 info s pos st st' =
       (* It is important to place these reads before the write *)
       (* But then, we don't have enough registers left to store the
          right position.  So, we store the position plus one. *)
-      let st'' = st'.next.(Char.code info.i_cols.[Char.code s.[pos]]) in
+      let st'' = st'.next.(Char.code (Bytes.get info.i_cols (Char.code s.[pos]))) in
       info.positions.(idx) <- pos;
       loop2 info s pos st' st''
     end else begin
@@ -219,7 +231,7 @@ and loop2 info s pos st st' =
 
 let rec loop_no_mark info s pos last st =
   if pos < last then
-    let st' = st.next.(Char.code info.i_cols.[Char.code s.[pos]]) in
+    let st' = st.next.(Char.code (Bytes.get info.i_cols (Char.code s.[pos]))) in
     let idx = st'.idx in
     if idx >= 0 then
       loop_no_mark info s (pos + 1) last st'
@@ -251,15 +263,13 @@ let find_initial_state re cat =
     re.initial_states <- (cat, st) :: re.initial_states;
     st
 
-let dummy_substrings = `Match ("", [], [||], 0)
-
-let get_color re s pos =
+let get_color re (s:string) pos =
   if pos < 0 then -1 else
   let slen = String.length s in
   if pos >= slen then -1 else
   (* Special case for the last newline *)
   if pos = slen - 1 && re.lnl <> -1 && s.[pos] = '\n' then re.lnl else
-  Char.code re.cols.[Char.code s.[pos]]
+  Char.code (Bytes.get re.cols (Char.code s.[pos]))
 
 let rec handle_last_newline info pos st groups =
   let st' = st.next.(info.re.lnl) in
@@ -272,7 +282,7 @@ let rec handle_last_newline info pos st groups =
     st'
   end else begin (* Unknown *)
     let c = info.re.lnl in
-    let real_c = Char.code info.i_cols.[Char.code '\n'] in
+    let real_c = Char.code (Bytes.get info.i_cols (Char.code '\n')) in
     let cat = category info.re c in
     let desc' = delta info cat real_c st in
     let st' = find_state info.re desc' in
@@ -280,14 +290,14 @@ let rec handle_last_newline info pos st groups =
     handle_last_newline info pos st groups
   end
 
-let rec scan_str info s initial_state groups =
+let rec scan_str info (s:string) initial_state groups =
   let pos = info.pos in
   let last = info.last in
   if
     last = String.length s &&
     info.re.lnl <> -1 &&
     last > pos &&
-    s.[last - 1] = '\n'
+    String.get s (last - 1) = '\n'
   then begin
     info.last <- last - 1;
     let st = scan_str info s initial_state groups in
@@ -300,7 +310,7 @@ let rec scan_str info s initial_state groups =
   else
     loop_no_mark info s pos last initial_state
 
-let match_str groups re s pos len =
+let match_str groups partial re s pos len =
   let slen = String.length s in
   let last = if len = -1 then slen else pos + len in
   let info =
@@ -323,7 +333,7 @@ let match_str groups re s pos len =
   let initial_state = find_initial_state re initial_cat in
   let st = scan_str info s initial_state groups in
   let res =
-    if st.idx = break then
+    if st.idx = break || partial then
       Automata.status st.desc
     else
       let final_cat =
@@ -336,10 +346,10 @@ let match_str groups re s pos len =
       res
   in
   match res with
-    `Match m ->
-      `Match (s, m, info.positions, re.group_count)
-  | (`Failed | `Running) as res ->
-      res
+    Automata.Match (marks, pmarks) ->
+      Match { s ; marks; pmarks ; gpos = info.positions; gcount = re.group_count}
+  | Automata.Failed -> Failed
+  | Automata.Running -> Running
 
 let mk_re init cols col_repr ncol lnl group_count =
   { initial = init;
@@ -360,13 +370,10 @@ let cseq c c' = Cset.seq (Char.code c) (Char.code c')
 let cadd c s = Cset.add (Char.code c) s
 let csingle c = Cset.single (Char.code c)
 
-let rec interval i j = if i > j then [] else i :: interval (i + 1) j
-
 let rec cset_hash_rec l =
   match l with
     []        -> 0
   | (i, j)::r -> i + 13 * j + 257 * cset_hash_rec r
-let cset_hash l = (cset_hash_rec l) land 0x3FFFFFFF
 
 module CSetMap =
   Map.Make
@@ -379,7 +386,7 @@ module CSetMap =
 let trans_set cache cm s =
   match s with
     [i, j] when i = j ->
-      csingle cm.[i]
+      csingle (Bytes.get cm i)
   | _ ->
       let v = (cset_hash_rec s, s) in
       try
@@ -387,7 +394,7 @@ let trans_set cache cm s =
       with Not_found ->
         let l =
           List.fold_right
-            (fun (i, j) l -> Cset.union (cseq cm.[i] cm.[j]) l)
+            (fun (i, j) l -> Cset.union (cseq (Bytes.get cm i) (Bytes.get cm j)) l)
             s Cset.empty
         in
         cache := CSetMap.add v l !cache;
@@ -395,8 +402,6 @@ let trans_set cache cm s =
 
 (****)
 
-type sem_status = Compulsory | Indicative
-
 type regexp =
     Set of Cset.t
   | Sequence of regexp list
@@ -413,6 +418,42 @@ type regexp =
   | Intersection of regexp list
   | Complement of regexp list
   | Difference of regexp * regexp
+  | Pmark of Automata.Pmark.t * regexp
+
+let rec pp fmt t =
+  let open Re_fmt in
+  let var s re = sexp fmt s pp re in
+  let seq s rel = sexp fmt s (list pp) rel in
+  match t with
+  | Set s ->  sexp fmt "Set" Cset.pp s
+  | Sequence sq -> seq "Sequence" sq
+  | Alternative alt -> seq "Alternative" alt
+  | Repeat (re, start, stop) ->
+    let pp' fmt () = fprintf fmt "%a@ %d%a" pp re   start   optint stop in
+    sexp fmt "Repeat" pp' ()
+  | Beg_of_line      -> str fmt "Beg_of_line"
+  | End_of_line      -> str fmt "End_of_line"
+  | Beg_of_word      -> str fmt "Beg_of_word"
+  | End_of_word      -> str fmt "End_of_word"
+  | Not_bound        -> str fmt "Not_bound"
+  | Beg_of_str       -> str fmt "Beg_of_str"
+  | End_of_str       -> str fmt "End_of_str"
+  | Last_end_of_line -> str fmt "Last_end_of_line"
+  | Start            -> str fmt "Start"
+  | Stop             -> str fmt "Stop"
+  | Sem (sem, re)    ->
+    sexp fmt "Sem" (pair Automata.pp_sem pp) (sem, re)
+  | Sem_greedy (k, re) ->
+    sexp fmt "Sem_greedy" (pair Automata.pp_rep_kind pp) (k, re)
+  | Group c        -> var "Group" c
+  | No_group c     -> var "No_group" c
+  | Nest c         -> var "Nest" c
+  | Case c         -> var "Case" c
+  | No_case c      -> var "No_case" c
+  | Intersection c -> seq "Intersection" c
+  | Complement c   -> seq "Complement" c
+  | Difference (a, b) -> sexp fmt "Difference" (pair pp pp) (a, b)
+  | Pmark (m, r)      -> sexp fmt "Pmark" (pair Automata.Pmark.pp pp) (m, r)
 
 let rec is_charset r =
   match r with
@@ -427,7 +468,8 @@ let rec is_charset r =
       is_charset r
   | Sequence _ | Repeat _ | Beg_of_line | End_of_line
   | Beg_of_word | End_of_word | Beg_of_str | End_of_str
-  | Not_bound | Last_end_of_line | Start | Stop | Group _ | Nest _ ->
+  | Not_bound | Last_end_of_line | Start | Stop
+  | Group _ | Nest _ | Pmark (_,_)->
       false
 
 (**** Colormap ****)
@@ -436,17 +478,20 @@ let rec is_charset r =
 let rec split s cm =
   match s with
     []    -> ()
-  | (i, j)::r -> cm.[i] <- '\001'; cm.[j + 1] <- '\001'; split r cm
+  | (i, j)::r -> Bytes.set cm i '\001'; Bytes.set cm (j + 1) '\001'; split r cm
 
 let cupper =
-  Cset.union (cseq 'A' 'Z') (Cset.union (cseq '\192' '\214') (cseq '\216' '\222'))
+  Cset.union (cseq 'A' 'Z')
+    (Cset.union (cseq '\192' '\214') (cseq '\216' '\222'))
 let clower = Cset.offset 32 cupper
-let calpha = cadd '\170' (cadd '\186' (Cset.union clower cupper))
+let calpha =
+  List.fold_right cadd ['\170'; '\181'; '\186'; '\223'; '\255']
+    (Cset.union clower cupper)
 let cdigit = cseq '0' '9'
 let calnum = Cset.union calpha cdigit
 let cword = cadd '_' calnum
 
-let rec colorize c regexp =
+let colorize c regexp =
   let lnl = ref false in
   let rec colorize regexp =
     match regexp with
@@ -458,12 +503,12 @@ let rec colorize c regexp =
     | Beg_of_word | End_of_word
     | Not_bound                 -> split cword c
     | Beg_of_str | End_of_str
-    | Start | Stop | Not_bound  -> ()
+    | Start | Stop              -> ()
     | Last_end_of_line          -> lnl := true
     | Sem (_, r)
     | Sem_greedy (_, r)
     | Group r | No_group r
-    | Nest r                    -> colorize r
+    | Nest r | Pmark (_,r)     -> colorize r
     | Case _ | No_case _
     | Intersection _
     | Complement _
@@ -472,23 +517,78 @@ let rec colorize c regexp =
   colorize regexp;
   !lnl
 
-let make_cmap () = String.make 257 '\000'
+let make_cmap () = Bytes.make 257 '\000'
 
 let flatten_cmap cm =
-  let c = String.create 256 in
-  let col_repr = String.create 256 in
+  let c = Bytes.create 256 in
+  let col_repr = Bytes.create 256 in
   let v = ref 0 in
-  c.[0] <- '\000';
-  col_repr.[0] <- '\000';
+  Bytes.set c 0 '\000';
+  Bytes.set col_repr 0 '\000';
   for i = 1 to 255 do
-    if cm.[i] <> '\000' then incr v;
-    c.[i] <- Char.chr !v;
-    col_repr.[!v] <- Char.chr i
+    if Bytes.get cm i <> '\000' then incr v;
+    Bytes.set c i (Char.chr !v);
+    Bytes.set col_repr !v (Char.chr i)
   done;
-  (c, String.sub col_repr 0 (!v + 1), !v + 1)
+  (c, Bytes.sub col_repr 0 (!v + 1), !v + 1)
 
 (**** Compilation ****)
 
+let rec equal x1 x2 =
+  match x1, x2 with
+    Set s1, Set s2 ->
+      s1 = s2
+  | Sequence l1, Sequence l2 ->
+      eq_list l1 l2
+  | Alternative l1, Alternative l2 ->
+      eq_list l1 l2
+  | Repeat (x1', i1, j1), Repeat (x2', i2, j2) ->
+      i1 = i2 && j1 = j2 && equal x1' x2'
+  | Beg_of_line, Beg_of_line
+  | End_of_line, End_of_line
+  | Beg_of_word, Beg_of_word
+  | End_of_word, End_of_word
+  | Not_bound, Not_bound
+  | Beg_of_str, Beg_of_str
+  | End_of_str, End_of_str
+  | Last_end_of_line, Last_end_of_line
+  | Start, Start
+  | Stop, Stop ->
+      true
+  | Sem (sem1, x1'), Sem (sem2, x2') ->
+      sem1 = sem2 && equal x1' x2'
+  | Sem_greedy (k1, x1'), Sem_greedy (k2, x2') ->
+      k1 = k2 && equal x1' x2'
+  | Group _, Group _ -> (* Do not merge groups! *)
+      false
+  | No_group x1', No_group x2' ->
+      equal x1' x2'
+  | Nest x1', Nest x2' ->
+      equal x1' x2'
+  | Case x1', Case x2' ->
+      equal x1' x2'
+  | No_case x1', No_case x2' ->
+      equal x1' x2'
+  | Intersection l1, Intersection l2 ->
+      eq_list l1 l2
+  | Complement l1, Complement l2 ->
+      eq_list l1 l2
+  | Difference (x1', x1''), Difference (x2', x2'') ->
+      equal x1' x2' && equal x1'' x2''
+  | Pmark (m1, r1), Pmark (m2, r2) ->
+      Automata.Pmark.equal m1 m2 && equal r1 r2
+  | _ ->
+      false
+
+and eq_list l1 l2 =
+  match l1, l2 with
+    [], [] ->
+      true
+  | x1 :: r1, x2 :: r2 ->
+      equal x1 x2 && eq_list r1 r2
+  | _ ->
+      false
+
 let sequence l =
   match l with
     [x] -> x
@@ -502,7 +602,7 @@ let rec merge_sequences l =
       merge_sequences (l' @ r)
   | Sequence (x :: y) :: r ->
       begin match merge_sequences r with
-        Sequence (x' :: y') :: r' when x = x' ->
+        Sequence (x' :: y') :: r' when equal x x' ->
           Sequence [x; Alternative [sequence y; sequence y']] :: r'
       | r' ->
           Sequence (x :: y) :: r'
@@ -519,7 +619,7 @@ let enforce_kind ids kind kind' cr =
   |  _               -> cr
 
 (* XXX should probably compute a category mask *)
-let rec translate ids kind ign_group ign_case greedy pos cache c r =
+let rec translate ids kind ign_group ign_case greedy pos cache (c:Bytes.t) r =
   match r with
     Set s ->
       (A.cst ids (trans_set cache c s), kind)
@@ -531,7 +631,7 @@ let rec translate ids kind ign_group ign_case greedy pos cache c r =
           let (cr, kind') =
             translate ids kind ign_group ign_case greedy pos cache c r' in
           (enforce_kind ids kind kind' cr, kind)
-      | l' ->
+      | _ ->
           (A.alt ids
              (List.map
                 (fun r' ->
@@ -628,6 +728,10 @@ let rec translate ids kind ign_group ign_case greedy pos cache c r =
         (A.seq ids `First (A.erase ids b e) cr, kind')
   | Difference _ | Complement _ | Intersection _ | No_case _ | Case _ ->
       assert false
+  | Pmark (i, r') ->
+    let (cr, kind') =
+      translate ids kind ign_group ign_case greedy pos cache c r' in
+    (A.seq ids `First (A.pmark ids i) cr, kind')
 
 and trans_seq ids kind ign_group ign_case greedy pos cache c l =
   match l with
@@ -713,6 +817,7 @@ let rec handle_case ign_case r =
   | Difference (r, r') ->
       Set (Cset.inter (as_set (handle_case ign_case r))
              (Cset.diff cany (as_set (handle_case ign_case r'))))
+  | Pmark (i,r) -> Pmark (i,handle_case ign_case r)
 
 (****)
 
@@ -734,6 +839,26 @@ let compile_1 regexp =
 
 (****)
 
+let rec anchored r =
+  match r with
+  | Sequence l ->
+      List.exists anchored l
+  | Alternative l ->
+      List.for_all anchored l
+  | Repeat (r, i, _) ->
+      i > 0 && anchored r
+  | Set _ | Beg_of_line | End_of_line | Beg_of_word | End_of_word
+  | Not_bound | End_of_str | Last_end_of_line | Stop
+  | Intersection _ | Complement _ | Difference _ ->
+      false
+  | Beg_of_str | Start ->
+      true
+  | Sem (_, r) | Sem_greedy (_, r) | Group r | No_group r | Nest r
+  | Case r | No_case r | Pmark (_, r) ->
+      anchored r
+
+(****)
+
 type t = regexp
 
 let str s =
@@ -769,6 +894,7 @@ let word r = seq [bow; r; eow]
 let not_boundary = Not_bound
 let bos = Beg_of_str
 let eos = End_of_str
+let whole_string r = seq [bos; r; eos]
 let leol = Last_end_of_line
 let start = Start
 let stop = Stop
@@ -780,6 +906,7 @@ let non_greedy r = Sem_greedy (`Non_greedy, r)
 let group r = Group r
 let no_group r = No_group r
 let nest r = Nest r
+let mark r = let i = Automata.Pmark.gen () in (i,Pmark (i,r))
 
 let set str =
   let s = ref [] in
@@ -813,6 +940,7 @@ let upper = alt [rg 'A' 'Z'; rg '\192' '\214'; rg '\216' '\222']
 let alpha = alt [lower; upper; char '\170'; char '\186']
 let digit = rg '0' '9'
 let alnum = alt [alpha; digit]
+let wordc = alt [alnum; char '_']
 let ascii = rg '\000' '\127'
 let blank = set "\t "
 let cntrl = alt [rg '\000' '\031'; rg '\127' '\159']
@@ -823,94 +951,313 @@ let punct =
        rg '\123' '\126'; rg '\160' '\169'; rg '\171' '\180';
        rg '\182' '\185'; rg '\187' '\191'; char '\215'; char '\247']
 let space = alt [char ' '; rg '\009' '\013']
-let xdigit = alt [digit; rg 'a' 'f'; rg 'A' 'Z']
+let xdigit = alt [digit; rg 'a' 'f'; rg 'A' 'F']
 
 let case r = Case r
 let no_case r = No_case r
 
 (****)
 
-type substrings = (string * Automata.mark_infos * int array * int)
+let compile r =
+  compile_1 (if anchored r then group r else seq [shortest (rep any); group r])
 
-let compile r = compile_1 (seq [shortest (rep any); group r])
-
-let exec ?(pos = 0) ?(len = -1) re s =
+let exec_internal name ?(pos=0) ?(len = -1) re s =
   if pos < 0 || len < -1 || pos + len > String.length s then
-    invalid_arg "Re.exec";
-  match match_str true re s pos len with
-    `Match substr -> substr
-  | _             -> raise Not_found
+    invalid_arg name;
+  match_str true false re s pos len
 
-let execp ?(pos = 0) ?(len = -1) re s =
-  if pos < 0 || len < -1 || pos + len > String.length s then
-    invalid_arg "Re.execp";
-  match match_str false re s pos len with
-    `Match substr -> true
+let exec ?pos ?len re s =
+  match exec_internal "Re.exec" ?pos ?len re s with
+    Match substr -> substr
+  | _            -> raise Not_found
+
+let exec_opt ?pos ?len re s =
+  match exec_internal "Re.exec_opt" ?pos ?len re s with
+    Match substr -> Some substr
+  | _            -> None
+
+let execp ?pos ?len re s =
+  match exec_internal "Re.execp" ?pos ?len re s with
+    Match _substr -> true
   | _             -> false
 
-let exec_partial ?(pos = 0) ?(len = -1) re s =
-  if pos < 0 || len < -1 || pos + len > String.length s then
-    invalid_arg "Re.exec_partial";
-  match match_str false re s pos len with
-    `Match _ -> `Full
-  | `Running -> `Partial
-  | `Failed  -> `Mismatch
+let exec_partial ?pos ?len re s =
+  match exec_internal "Re.exec_partial" ?pos ?len re s with
+    Match _ -> `Full
+  | Running -> `Partial
+  | Failed  -> `Mismatch
 
-let rec find_mark (i : int) l =
-  match l with
-    [] ->
-      raise Not_found
-  | (j, idx) :: r ->
-      if i = j then idx else find_mark i r
+module Group = struct
 
-let get (s, marks, pos, _) i =
-  if 2 * i + 1 >= Array.length marks then raise Not_found;
-  let m1 = marks.(2 * i) in
-  if m1 = -1 then raise Not_found;
-  let p1 = pos.(m1) - 1 in
-  let p2 = pos.(marks.(2 * i + 1)) - 1 in
-  String.sub s p1 (p2 - p1)
+  type t = groups
 
-let get_ofs (s, marks, pos, _) i =
-  if 2 * i + 1 >= Array.length marks then raise Not_found;
-  let m1 = marks.(2 * i) in
-  if m1 = -1 then raise Not_found;
-  let p1 = pos.(m1) - 1 in
-  let p2 = pos.(marks.(2 * i + 1)) - 1 in
-  (p1, p2)
-
-let test (s, marks, pos, _) i =
-  if 2 * i >= Array.length marks then false else
-  let idx = marks.(2 * i) in
-  idx <> -1
-
-let dummy_offset = (-1, -1)
-
-let get_all_ofs ((s, marks, pos, count) as m) =
-  let res = Array.make count dummy_offset in
-  for i = 0 to Array.length marks / 2 - 1 do
+  let get {s ; marks ; gpos ; _} i =
+    if 2 * i + 1 >= Array.length marks then raise Not_found;
     let m1 = marks.(2 * i) in
-    if m1 <> -1 then begin
-      let p1 = pos.(m1) in
-      let p2 = pos.(marks.(2 * i + 1)) in
-      res.(i) <- (p1 - 1, p2 - 1)
-    end
-  done;
-  res
+    if m1 = -1 then raise Not_found;
+    let p1 = gpos.(m1) - 1 in
+    let p2 = gpos.(marks.(2 * i + 1)) - 1 in
+    String.sub s p1 (p2 - p1)
 
-let dummy_string = ""
-
-let get_all ((s, marks, pos, count) as m) =
-  let res = Array.make count dummy_string in
-  for i = 0 to Array.length marks / 2 - 1 do
+  let offset {marks ; gpos ; _} i =
+    if 2 * i + 1 >= Array.length marks then raise Not_found;
     let m1 = marks.(2 * i) in
-    if m1 <> -1 then begin
-      let p1 = pos.(m1) in
-      let p2 = pos.(marks.(2 * i + 1)) in
-      res.(i) <- String.sub s (p1 - 1) (p2 - p1)
+    if m1 = -1 then raise Not_found;
+    let p1 = gpos.(m1) - 1 in
+    let p2 = gpos.(marks.(2 * i + 1)) - 1 in
+    (p1, p2)
+
+  let start subs i = fst (offset subs i)
+
+  let stop subs i = snd (offset subs i)
+
+  let test { marks ; _ } i =
+    if 2 * i >= Array.length marks then false else
+      let idx = marks.(2 * i) in
+      idx <> -1
+
+  let dummy_offset = (-1, -1)
+
+  let all_offset {marks ; gpos ; gcount ; _} =
+    let res = Array.make gcount dummy_offset in
+    for i = 0 to Array.length marks / 2 - 1 do
+      let m1 = marks.(2 * i) in
+      if m1 <> -1 then begin
+        let p1 = gpos.(m1) in
+        let p2 = gpos.(marks.(2 * i + 1)) in
+        res.(i) <- (p1 - 1, p2 - 1)
+      end
+    done;
+    res
+
+  let dummy_string = ""
+
+  let all {s ; marks ; gpos ; gcount ; _ } =
+    let res = Array.make gcount dummy_string in
+    for i = 0 to Array.length marks / 2 - 1 do
+      let m1 = marks.(2 * i) in
+      if m1 <> -1 then begin
+        let p1 = gpos.(m1) in
+        let p2 = gpos.(marks.(2 * i + 1)) in
+        res.(i) <- String.sub s (p1 - 1) (p2 - p1)
+      end
+    done;
+    res
+
+  let pp fmt t =
+    let matches =
+      let offsets = all_offset t in
+      let strs = all t in
+      Array.to_list (
+        Array.init (Array.length strs) (fun i -> strs.(i), offsets.(i))
+      ) in
+    let open Re_fmt in
+    let pp_match fmt (str, (start, stop)) =
+      fprintf fmt "@[(%s (%d %d))@]" str start stop in
+    sexp fmt "Group" (list pp_match) matches
+end
+
+module Mark = struct
+
+  type t = Automata.Pmark.t
+
+  let test {pmarks ; _} p =
+    Automata.PmarkSet.mem p pmarks
+
+  let all s = s.pmarks
+
+  module Set = MarkSet
+
+  let equal = Automata.Pmark.equal
+
+  let compare = Automata.Pmark.compare
+
+end
+
+type 'a gen = unit -> 'a option
+
+let all_gen ?(pos=0) ?len re s =
+  if pos < 0 then invalid_arg "Re.all";
+  (* index of the first position we do not consider.
+    !pos < limit is an invariant *)
+  let limit = match len with
+    | None -> String.length s
+    | Some l ->
+        if l<0 || pos+l > String.length s then invalid_arg "Re.all";
+        pos+l
+  in
+  (* iterate on matches. When a match is found, search for the next
+     one just after its end *)
+  let pos = ref pos in
+  fun () ->
+    if !pos >= limit
+    then None  (* no more matches *)
+    else match match_str true false re s !pos (limit - !pos) with
+      | Match substr ->
+          let p1, p2 = Group.offset substr 0 in
+          pos := if p1=p2 then p2+1 else p2;
+          Some substr
+      | Running
+      | Failed -> None
+
+let all ?pos ?len re s =
+  let l = ref [] in
+  let g = all_gen ?pos ?len re s in
+  let rec iter () = match g() with
+    | None -> List.rev !l
+    | Some sub -> l := sub :: !l; iter ()
+  in iter ()
+
+let matches_gen ?pos ?len re s =
+  let g = all_gen ?pos ?len re s in
+  fun () ->
+    match g() with
+    | None -> None
+    | Some sub -> Some (Group.get sub 0)
+
+let matches ?pos ?len re s =
+  let l = ref [] in
+  let g = all_gen ?pos ?len re s in
+  let rec iter () = match g() with
+    | None -> List.rev !l
+    | Some sub -> l := Group.get sub 0 :: !l; iter ()
+  in iter ()
+
+type split_token =
+  [ `Text of string
+  | `Delim of groups
+  ]
+
+let split_full_gen ?(pos=0) ?len re s =
+  if pos < 0 then invalid_arg "Re.split";
+  let limit = match len with
+    | None -> String.length s
+    | Some l ->
+        if l<0 || pos+l > String.length s then invalid_arg "Re.split";
+        pos+l
+  in
+  (* i: start of delimited string
+    pos: first position after last match of [re]
+    limit: first index we ignore (!pos < limit is an invariant) *)
+  let pos0 = pos in
+  let state = ref `Idle in
+  let i = ref pos and pos = ref pos in
+  let next () = match !state with
+  | `Idle when !pos >= limit ->
+      if !i < limit then (
+        let sub = String.sub s !i (limit - !i) in
+        incr i;
+        Some (`Text sub)
+      ) else None
+  | `Idle ->
+    begin match match_str true false re s !pos (limit - !pos) with
+      | Match substr ->
+          let p1, p2 = Group.offset substr 0 in
+          pos := if p1=p2 then p2+1 else p2;
+          let old_i = !i in
+          i := p2;
+          if p1 > pos0 then (
+            (* string does not start by a delimiter *)
+            let text = String.sub s old_i (p1 - old_i) in
+            state := `Yield (`Delim substr);
+            Some (`Text text)
+          ) else Some (`Delim substr)
+      | Running -> None
+      | Failed ->
+          if !i < limit
+          then (
+            let text = String.sub s !i (limit - !i) in
+            i := limit;
+            Some (`Text text)  (* yield last string *)
+          ) else None
     end
-  done;
-  res
+  | `Yield x ->
+      state := `Idle;
+      Some x
+  in next
+
+let split_full ?pos ?len re s =
+  let l = ref [] in
+  let g = split_full_gen ?pos ?len re s in
+  let rec iter () = match g() with
+    | None -> List.rev !l
+    | Some s -> l := s :: !l; iter ()
+  in iter ()
+
+let split_gen ?pos ?len re s =
+  let g = split_full_gen ?pos ?len re s in
+  let rec next() = match g()  with
+    | None -> None
+    | Some (`Delim _) -> next()
+    | Some (`Text s) -> Some s
+  in next
+
+let split ?pos ?len re s =
+  let l = ref [] in
+  let g = split_full_gen ?pos ?len re s in
+  let rec iter () = match g() with
+    | None -> List.rev !l
+    | Some (`Delim _) -> iter()
+    | Some (`Text s) -> l := s :: !l; iter ()
+  in iter ()
+
+let replace ?(pos=0) ?len ?(all=true) re ~f s =
+  if pos < 0 then invalid_arg "Re.replace";
+  let limit = match len with
+    | None -> String.length s
+    | Some l ->
+        if l<0 || pos+l > String.length s then invalid_arg "Re.replace";
+        pos+l
+  in
+  (* buffer into which we write the result *)
+  let buf = Buffer.create (String.length s) in
+  (* iterate on matched substrings. *)
+  let rec iter pos =
+    if pos < limit
+    then match match_str true false re s pos (limit-pos) with
+      | Match substr ->
+          let p1, p2 = Group.offset substr 0 in
+          (* add string between previous match and current match *)
+          Buffer.add_substring buf s pos (p1-pos);
+          (* what should we replace the matched group with? *)
+          let replacing = f substr in
+          Buffer.add_string buf replacing;
+          if all
+          (* if we matched a non-char e.g. ^ we must manually advance by 1 *)
+          then iter
+                 (if p1=p2
+                  then (
+                    (* a non char could be past the end of string. e.g. $ *)
+                    if p2 < limit then Buffer.add_char buf s.[p2];
+                    p2+1)
+                  else p2)
+          else Buffer.add_substring buf s p2 (limit-p2)
+      | Running -> ()
+      | Failed ->
+          Buffer.add_substring buf s pos (limit-pos)
+  in
+  iter pos;
+  Buffer.contents buf
+
+let replace_string ?pos ?len ?all re ~by s =
+  replace ?pos ?len ?all re s ~f:(fun _ -> by)
+
+
+
+(** {2 Deprecated functions} *)
+
+type substrings = groups
+
+let get = Group.get
+let get_ofs = Group.offset
+let get_all = Group.all
+let get_all_ofs = Group.all_offset
+let test = Group.test
+
+type markid = Mark.t
+
+let marked = Mark.test
+let mark_set = Mark.all
 
 (**********************************)
 
